#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import block
from arrlp import get_xp, sortloop
import bottleneck as bn
import numpy as np
from scipy.stats import linregress



# %% Function
@block()
def detect_gain(channels, /, *, nbins=50, cuda=False, parallel=False) :
    '''
    This function calculates the gain maps.
    '''

    xp = get_xp(cuda)
    gain_maps, mean_maps, var_maps = [], [], []
    mean_bins, var_bins = [], []
    gains, fits = [], []

    for channel in channels :
        channel = xp.asarray(channel)
        med = xp.median(channel, axis=0) if cuda else bn.median(channel, axis=0)
        mad = xp.median(xp.abs(channel-med), axis=0) if cuda else bn.median(xp.abs(channel-med), axis=0)
        mean = med
        var = (1.4826 * mad)**2
        if cuda: mean, var = xp.asnumpy(mean), xp.asnumpy(var)
        gain_maps.append(mean / var)
        mean_maps.append(mean)
        var_maps.append(var)
        
        # Bin
        bins = np.linspace(mean.min(), mean.max(), nbins+1)
        digitized = np.digitize(mean, bins)
        mean_binned = []
        var_binned = []
        for _, _, mean_m, var_m in sortloop(digitized, mean, var) :
            if len(mean_m) > 10:
                mean_binned.append(mean_m.mean())
                var_binned.append(var_m.mean())
        mean_binned, var_binned = np.array(mean_binned), np.array(var_binned)
        mean_bins.append(mean_binned)
        var_bins.append(var_binned)

        # Fit
        mask = np.logical_and(mean_binned < np.percentile(mean, 66), mean_binned > np.percentile(mean, 33))
        fit = linregress(mean_binned[mask], (var_binned[mask]))
        gain = 1.0 / fit.slope
        gains.append(gain)
        fits.append(fit)

    return gains, dict(gain=gain_maps, mean=mean_maps, var=var_maps, mean_bins=mean_bins, var_bins=var_bins, fit=fits)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import columns
from warnings import warn



def open_df(locs, df) :

    # Get dataframe instance
    index = df.index.name
    if index is not None :
        index = index.replace('"', '')
        index = index.replace("'", "")
    else :
        index = 'detection'
    if index not in columns.headers :
        warn(f'Skipping opening unknown dataframe with index "{index}"')
        return None
    col = columns.headers[index]
    df_name = col.df_name
    dataframe =  getattr(locs, df_name)

    # Apply column
    for header in df.columns :
        if header not in columns.headers :
            warn(f'Skipping opening unknown column with header "{header}"')
        col = columns.headers[header]
        col_name = col.col
        dtype = col.dtype
        array = df[header].to_numpy().astype(dtype)
        setattr(dataframe, col_name, array)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import block, Config
from arrlp import gc, get_xp, img_autocorr, img_fft, img_ifft, coordinates
import bottleneck as bn
from funclp import Gaussian2D, Spline2D
import numpy as np
from scipy.optimize import curve_fit



# %% Function
@block()
def blink_spatial_psf(channels, /, crop=20, *,  channel_pixel=1., cuda=False, parallel=False) :
    '''
    This function creates the global psf image and characterizations from the channels spatial autocorrelations.
    '''

    # Optimization parameters
    xp = get_xp(cuda)

    # Get pixel
    pixel = Config(nfiles=len(channels), pixel=channel_pixel).pixel

    # Coordinates
    Y, X = coordinates((2*crop+1, 2*crop+1), grid=True)
    mask = np.logical_and(Y != 0, X != 0)

    psf_sigma = []
    results = {'ac': [], 'psf':[], 'fit': [], 'spline': [], 'sigx': [], 'sigy': [], 'theta': [], 'tx': [], 'ty': [], 'coeffs': []}
    for channel, pix in zip(channels, pixel) :

        # Calculating autocorrelation
        gc()
        bkgd = xp.median(channel, axis=(0,), keepdims=True) if cuda else bn.median(channel, axis=0).reshape((1, *channel.shape[1:]))
        bkgd = xp.minimum(bkgd, channel)
        channel = channel - bkgd
        y0, x0 = int(channel.shape[1]//2), int(channel.shape[2]//2)
        ac = img_autocorr(channel, stacks=True, cuda=cuda, parallel=parallel)
        ac = ac[:, y0 - crop : y0 + crop + 1, x0 - crop : x0 + crop + 1]
        ac = ac.mean(axis=0)
        if cuda : ac = xp.asnumpy(ac)
        mid, ac[crop, crop] = ac[crop, crop], np.nan
        ac -= np.nanmin(ac)
        maxi = np.nanmax(ac)
        ac[crop, crop] = mid
        ac /= maxi

        # Calculating PSF
        gc()
        psf = np.fft.fftshift(np.real(img_ifft(np.sqrt(np.abs(img_fft(ac))))))
        mid, psf[crop, crop] = psf[crop, crop], np.nan
        psf -= np.nanmin(psf)
        maxi = np.nanmax(psf)
        psf[crop, crop] = mid
        psf /= maxi

        # Fitting
        gc()
        _psf = psf[mask]
        yy = Y[mask] * pix[0] / min(pix)
        xx = X[mask] * pix[1] / min(pix)
        p0 = [1., 1., 1., 0., 0.] # sigx, sigy, amp, offset, theta [°]
        bounds = ([0., 0., 0.75, -0.25, 0.], [crop, crop, 1.25, 0.25, 90.]) # (lower, upper)
        gaus = Gaussian2D(pixx=1., pixy=1., sigx=2., sigy=2.)
        func2fit = lambda xy, sigx, sigy, amp, offset, theta: gaus(xy[0], xy[1], sigx=sigx, sigy=sigy, amp=amp, offset=offset, theta=theta)
        popt, _ = curve_fit(func2fit, (xx, yy), _psf, p0=p0, bounds=bounds)
        sigx, sigy, amp, offset, theta = popt
        sigx, sigy = sigx * min(pix), sigy * min(pix)
        _X, _Y = X * pix[1] / min(pix), Y * pix[0] / min(pix)
        fit = func2fit((_X, _Y), *popt)
        maxi = fit.max()
        fit = (fit-offset) / amp
        psf = (psf-offset) / amp
        
        # Spline
        yy = Y * pix[0]
        xx = X * pix[1]
        mask = np.ones(psf.shape, dtype=bool)
        mask[crop, :] = False
        mask[:, crop] = False
        _psf = psf[mask].reshape(2*crop, 2*crop)
        _yy = yy[mask].reshape(2*crop, 2*crop)
        _xx = xx[mask].reshape(2*crop, 2*crop)
        spline_model = Spline2D(_psf, _xx, _yy)
        spline = spline_model(xx, yy)



        # Appending to results
        psf_sigma.append(np.sqrt(sigx * sigy))
        results['ac'].append(ac)
        results['psf'].append(psf)
        results['fit'].append(fit)
        results['spline'].append(spline)
        results['sigx'].append(sigx)
        results['sigy'].append(sigy)
        results['theta'].append(theta)
        results['tx'].append(spline_model.tx)
        results['ty'].append(spline_model.ty)
        results['coeffs'].append(spline_model.coeffs)

    return psf_sigma, results


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-03-03
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : smlmLP
# Module        : computer

"""
This object gives the various computer current parameters.
"""



# %% Libraries
import psutil
import shutil
import numba.cuda
import subprocess
from arrlp import gc



# %% Class
class Computer() :
    def __init__(self) :
        self.ram = RAM()
        self.vram = VRAM()
        self.cpu = CPU()
        self.gpu = GPU()
        self.disc = Disc()

class RAM() :
    def total(self) :
        gc()
        return psutil.virtual_memory().total / (1024 ** 3)
    def free(self) :
        gc()
        return psutil.virtual_memory().available / (1024 ** 3)
    def used(self) :
        gc()
        return psutil.virtual_memory().used / (1024 ** 3)

class VRAM() :
    def total(self) :
        if not numba.cuda.is_available() : return 0
        gc()
        info = subprocess.run(['nvidia-smi', '--query-gpu=name,memory.total,memory.free,memory.used,utilization.gpu,temperature.gpu', '--format=csv,noheader'], capture_output=True, text=True).stdout.strip().split("\n")[0].split(", ")
        return float(info[1].split(' ')[0]) / 1000 
    def free(self) :
        if not numba.cuda.is_available() : return 0
        gc()
        info = subprocess.run(['nvidia-smi', '--query-gpu=name,memory.total,memory.free,memory.used,utilization.gpu,temperature.gpu', '--format=csv,noheader'], capture_output=True, text=True).stdout.strip().split("\n")[0].split(", ")
        return float(info[2].split(' ')[0]) / 1000
    def used(self) :
        if not numba.cuda.is_available() : return 0
        gc()
        info = subprocess.run(['nvidia-smi', '--query-gpu=name,memory.total,memory.free,memory.used,utilization.gpu,temperature.gpu', '--format=csv,noheader'], capture_output=True, text=True).stdout.strip().split("\n")[0].split(", ")
        return float(info[3].split(' ')[0]) / 1000

class CPU() :
    def cores(self) :
        return psutil.cpu_count(logical=True)

class GPU() :
    def cuda(self) :
        return numba.cuda.is_available()

class Disc() :
    def total(self) :
        return shutil.disk_usage('/').total / (1024 ** 3)
    def free(self) :
        return shutil.disk_usage('/').free / (1024 ** 3)
    def used(self) :
        return shutil.disk_usage('/').used / (1024 ** 3)

computer = Computer()



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import block, Config
from arrlp import img_greyopening, kernel
import numpy as np



# %% Function
@block()
def bkgd_spatial_opening(channels, /, opening_radius=5, bkgds=None, noise_corrections=None, *, channel_pixel=1., cuda=False, parallel=False) :
    '''
    This function creates the spatial opening background.
    '''

    # Get local sigma in pixel
    channel_pixel = Config(nfiles=len(channels), pixel=channel_pixel).pixel

    # Correct bkgd length for end of acquisition
    if bkgds is not None and len(bkgds[0]) > len(channels[0]):
        bkgds = [bkgd[:len(channel)] for channel, bkgd in zip(channels, bkgds)]
    
    # Noise corrections
    if noise_corrections is None :
        noise_corrections = [np.float32(1.) for _ in range(len(channels))]

    #Footprint
    footprints = [kernel(window=2*opening_radius, pixel=pix) for pix in channel_pixel]

    new_bkgds = []
    for i in range(len(channels)) :
        bkgd = None if bkgds is None else bkgds[i]
        channel = channels[i]
        new_bkgd = img_greyopening(channel, footprint=footprints[i], out=bkgd, cuda=cuda, parallel=parallel, stacks=True)
        noise_corrections[i] *= np.float32(1.)
        new_bkgds.append(new_bkgd)

    return new_bkgds, noise_corrections


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import block, Config
from arrlp import get_xp, nb_threads
import numba as nb
from numba import cuda as nb_cuda 
import math



# %% Function
@block()
def detect_snr(signals, bkgds, noise_corrections=None, channel_gain=0.25, *, cuda=False, parallel=False) :
    '''
    This function normalizes the signals into SNRs.
    '''

    # xp
    xp = get_xp(cuda)

    # lists
    gains = Config(nfiles=len(signals), gain_experimental=channel_gain).gain
    if noise_corrections is None :
        noise_corrections = [xp.float32(1.) for _ in range(len(signals))]

    snrs = []
    for i in range(len(signals)) :
        signal = xp.asarray(signals[i], dtype=xp.float32)
        bkgd = xp.asarray(bkgds[i], dtype=xp.float32)
        noise_correction = xp.float32(noise_corrections[i])
        gain = xp.float32(gains[i])

        # Calculating SNR
        if cuda :
            threads_per_block = 256
            blocks_per_grid = (signal.size + threads_per_block - 1) // threads_per_block
            snr_gpu[blocks_per_grid, threads_per_block](signal.ravel(), bkgd.ravel(), noise_correction, gain)
        else :
            with nb_threads(parallel) :
                snr_cpu(signal.ravel(), bkgd.ravel(), noise_correction, gain)



@nb.njit(parallel=True, nogil=True, cache=True, fastmath=True)
def snr_cpu(signal, bkgd, noise_correction, gain) :
    for i in nb.prange(len(signal)) :
        signal[i] = nb.float32(signal[i] / noise_correction / math.sqrt(bkgd[i] / gain))



@nb_cuda.jit(fastmath=True)
def snr_gpu(signal, bkgd, noise_correction, gain) :
    i = nb_cuda.grid(1)
    if i < len(signal) :
        signal[i] = nb.float32(signal[i] / noise_correction / math.sqrt(bkgd[i] / gain))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-25
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : smlmLP
# Module        : Config

"""
This class stores configuration values in an object that can be saved and loaded.
"""



# %% Libraries
from corelp import prop, selfkwargs
from pathlib import Path
import json
from smlmlp import metadatum
from contextlib import ExitStack
import tifffile as tiff
from stacklp import shapetif



# %% Class
class Config() :
    '''
    This class stores configuration values in an object that can be saved and loaded.
    
    Parameters
    ----------
    a : int or float
        TODO.

    Attributes
    ----------
    _attr : int or float
        TODO.

    Examples
    --------
    >>> from smlmlp import Config
    ...
    >>> instance = Config(TODO)
    '''


    # Init
    def __init__(self, config=None, **kwargs) :

        if config is not None :

            # Open file
            if isinstance(config, str) or isinstance(config,Path('').__class__):
                config_file = Path(config).with_suffix('.json')
                if config_file.exists() :
                    with open(config_file, "r") as file:
                        config = json.load(file)
                else :
                    raise SyntaxError(f'config path was not recognized: {config}')
            selfkwargs(self, config)
        selfkwargs(self, kwargs)

    @property
    def metadata(self) :
        return {datum : getattr(self, datum, None) for group in metadatum.groups for datum in group}

    def save(self, path, file=None) :
        path = Path(path)
        if file is not None : path = path / file
        path = path.with_suffix('.json')
        with open(path, 'w') as json_file:
            json.dump(self.metadata, json_file, indent=4)

    def file_update(self, *tif_paths) :
        with ExitStack() as stack :
            tifs = [stack.enter_context(tiff.TiffFile(file)) for file in tif_paths]
            shapes = [shapetif(tif) for tif in tifs]

            # Check number of frames
            nframes = None
            for shape in shapes :
                if nframes is None : nframes = shape[0]
                if shape[0] != nframes :
                    raise ValueError('All tiff files do not have the same number of frames which is not possible for a single SMLM acquisition')
            self.nframes = nframes

            # Define shape
            # TODO




    # --- Below are the metadata, organized by group---



    # Cameras
    



    # Image size
    @property
    def shape(self) :
        return (self.yshape, self.xshape)
    @shape.setter
    def shape(self, value) :
        self.yshape, self.xshape = int(value), int(value)
    @metadatum("Image size", unit='pixel', ui=False)
    def xshape(self) :
        return 1024
    @metadatum("Image size", unit='pixel', ui=False)
    def yshape(self) :
        return 1024
    @metadatum("Image size", unit='µm', ui=False)
    def xFOV(self) : # µm
        return self.xshape * self.xpixel * 1e-3
    @metadatum("Image size", unit='µm', ui=False)
    def yFOV(self) : # µm
        return self.yshape * self.xpixel * 1e-3
    

    # 
    @metadatum(unit='ypix, xpix')
    def cropshape(self) :
        return (int(round(self.PSF_field/self.ypixel)), int(round(self.PSF_field/self.xpixel)))



    # PSF
    @metadatum('PSF', unit='nm')
    def wavelength(self) : # Fluorescence wavelength [nm]
        return 671.
    @metadatum('PSF')
    def NA(self) : # Objective Numerical Aperture
        return 1.5
    @metadatum('PSF', ui=['air', 'water', 'glycerol', 'silicon', 'oil'])
    def immersion_medium(self) : # Optical_index of immersion_medium
        return 'oil'
    @metadatum('PSF', ui=False)
    def optical_index(self) :
        dic = {'air': 1., 'water': 1.33, 'glycerol': 1.4, 'silicon': 1.4, 'oil': 1.515}
        return dic[self.immersion_medium]
    @metadatum('PSF')
    def PSF_fact(self) : # Experimental PSF growth factor
        return 2.
    @property
    def PSF_sigma(self) : #PSF gaussian sigma [nm]
        return 0.21 * self.wavelength / self.NA * self.PSF_fact
    @property
    def PSF_field(self) :
        return 1.22 * self.wavelength / self.NA * self.PSF_fact
    @property
    def DOF(self) : #Depth of Field [nm]
        delta_wave = self.immersion_medium * self.wavelength / self.NA**2
        delta_geom = self.immersion_medium * self.pixel / self.NA
        return delta_wave + delta_geom



    #Photon count
    @metadatum("Photon count")
    def ADU(self) : # Analog Digital Unit of camera
        return 0.25
    @metadatum("Photon count")
    def QE(self) : # Quantum efficiency of camera at wavelength
        return 0.72
    @metadatum("Photon count")
    def camera_gain(self) : # Camera gain
        return 1.
    @metadatum("Photon count")
    def camera_compensation(self) : # Camera readout noise compensation
        return 0.



    #Density
    @metadatum("Localization density", unit='loc/µm²')
    def density(self) : # density of localization [loc/µm²]
        return 0.1
    @metadatum("Localization density", unit='nm', ui=False)
    def density_radius(self) :
        return np.sqrt(1 / self.density / np.pi) * 1000 # [nm]
    @density_radius.setter
    def density_radius(self, value) :
        self.density = 1 / (value / 1000)**2 / np.pi
    @property
    def PSF_influence(self) : # mean radius of influence of a PSF [nm]
        return np.exp(((np.log(3 * self.PSF_sigma) + np.log(self.density_radius)) / 2)) # logarithmic average



    #Photophysics
    @metadatum("Photophysics", unit='ms')
    def exposure(self) : # camera exposure time [ms]
        return 50. 
    @metadatum("Photophysics", ui=['STORM', 'PAINT', 'PALM'])
    def imaging_modality(self) :
        return 'STORM'
    @metadatum("Photophysics", unit='ms')
    def tau_ON(self) : # Mean ON time of blinking events
        dic = dict(STORM=15., PAINT=200., PALM=15.)
        if self.imaging_modality in dic : return dic[self.imaging_modality]
        return self.exposure



    # Defines number of loops and hom many frames to charge
    @property
    def frame_bytes(self) : # gigabytes/frame
        return np.sum([(size[0] * size[1] * dtype.itemsize / 1024 ** 3) for size, dtype in zip(self.sizes,self.dtypes)])
    chunk_fraction = 1/25 # fraction of available ram to define how many frames to load at once
    @property
    def chunk_in_memory(self):
        return int(self.memory*self.chunk_fraction // self.frame_bytes) +1
    pad = 0
    @metadatum(unit='frames')
    def load(self) : # number of frames loaded in stack
        if getattr(self,'_load',None) is not None :
            return max(1,min(self.nframes+2*self.pad,self._load))
        elif getattr(self,'_chunk',None) is not None :
            return self._chunk + 2*self.pad
        return self.chunk + 2*self.pad
    @load.setter
    def load(self,value) :
        self._load = value
        self._chunk = None
    @metadatum(unit='frames')
    def chunk(self) : # number of frames to process at each loop
        if getattr(self,'_chunk',None) is not None :
            return max(1,min(self.nframes,self._chunk))
        elif getattr(self,'_load',None) is not None :
            return self._load - 2*self.pad
        return min(self.nframes, self.chunk_in_memory)
    @chunk.setter
    def chunk(self,value) :
        self._chunk = value
        self._load = None
    @property
    def nloops(self) :
        return int(np.ceil(self.nframes/self.chunk))



    # Defines how to divide into channels
    _nchannels = None #Number of channels
    @metadatum(unit='channels')
    def nchannels(self) :
        if self._channel2load is not None :
            return len(self._channel2load) #If channels defined takes number directly
        return self.nfiles #If not defined, 1 channel / load

    _channel2load = None
    @metadatum(unit='n°load')
    def channel2load(self) : #For each channel, load index in which to get it
        if self.nfiles == 1 : #If only 1 file
            return list(iterable(0, self.nchannels))
        return list(np.arange(self.nfiles))
    @property
    def load2channels(self) : #For each load, gives channel indices
        channel2load = self.channel2load
        n = np.max(channel2load)+1
        return [[ch for ch, ld in enumerate(channel2load) if load==ld] for load in range(n)]
    @property
    def channel2channel(self) :
        return [[ch] for ch in range(self.nchannels)]

    @metadatum(unit='(x0,y0,x1,y1)')
    def channel_bbox(self) : #(x0, y0, x1, y1)
        if self.nchannels == self.nfiles : #Takes full frame
            return [(0, 0, self.sizes[load][0], self.sizes[load][1]) for load in self.channel2load]
        if self.nchannels == 1 : #Takes full frame
            return [(0, 0, self.sizes[0][0], self.sizes[0][1])]
        if self.nchannels == 2 and self.nfiles == 1: # Divides along x, sCMOS rapid direction
            return [
                    (0, 0, self.sizes[0][0]//2, self.sizes[0][1]),
                    (self.sizes[0][0] - self.sizes[0][0]//2, 0, self.sizes[0][0], self.sizes[0][1]),
                    ]
        if self.nchannels == 4 and self.nfiles == 1: # Divides along x and y, sCMOS rapid direction
            return [
                    (0, 0, self.sizes[0][0]//2, self.sizes[0][1]//2),
                    (self.sizes[0][0] - self.sizes[0][0]//2, 0, self.sizes[0][0], self.sizes[0][1]//2),
                    (0, self.sizes[0][1] - self.sizes[0][1]//2, self.sizes[0][0]//2, self.sizes[0][1]),
                    (self.sizes[0][0] - self.sizes[0][0]//2, self.sizes[0][1] - self.sizes[0][1]//2, self.sizes[0][0], self.sizes[0][1]),
                    ]
        if self.nchannels == 4 and self.nfiles == 2: # Divides along x on the two files, sCMOS rapid direction
            return [
                    (0, 0, self.sizes[0][0]//2, self.sizes[0][1]),
                    (self.sizes[0][0] - self.sizes[0][0]//2, 0, self.sizes[0][0], self.sizes[0][1]),
                    (0, 0, self.sizes[1][0]//2, self.sizes[1][1]),
                    (self.sizes[1][0] - self.sizes[1][0]//2, 0, self.sizes[1][0], self.sizes[1][1]),
                    ]
        if self.nchannels == 5 and self.nfiles == 2 :
            return [
                    (0, 0, self.sizes[0][0]//2, self.sizes[0][1]//2),
                    (self.sizes[0][0] - self.sizes[0][0]//2, 0, self.sizses[0][0], self.sizes[0][1]//2),
                    (0, self.sizes[0][1] - self.sizes[0][1]//2, self.sizes[0][0]//2, self.sizes[0][1]),
                    (self.sizes[0][0] - self.sizes[0][0]//2, self.sizes[0][1] - self.sizes[0][1]//2, self.sizes[0][0], self.sizes[0][1]),
                    (0, 0, self.sizes[1][0], self.sizes[1][1])
                    ]
        return [(0, 0, self.sizes[0][0], self.sizes[0][1]) for _ in self.channel2load] #If nothing, just gives first full frame n times



    #Channel info
    @property
    def channel_shapes(self) :
        return [(y1-y0, x1-x0) for x0, y0, x1, y1 in self.channel_bbox]
    @property
    def channel_sizes(self) :
        return [(shape[1], shape[0]) for shape in self.channel_shapes]
    @property
    def channel_xpixels(self) :
        return [self.xpixels[load] for load in self.channel2load]
    @property
    def channel_ypixels(self) :
        return [self.ypixels[load] for load in self.channel2load]
    @property
    def channel_pixels(self) :
        return [np.sqrt(xpix*ypix) for xpix, ypix in zip(self.channel_xpixels, self.channel_ypixels)]
    @metadatum(unit='xpix, ypix')
    def channel_cropsizes(self) :
        return [(np.uint8(round(self.PSF_field/xpix)),np.uint8(round(self.PSF_field/ypix))) for xpix, ypix in zip(self.channel_xpixels, self.channel_ypixels)]
    @property
    def channel_cropshapes(self) :
        return [(size[1], size[0]) for size in self.channel_cropsizes]



    #Translations
    @metadatum(unit='nm')
    def channel_shifts(self) : #[(shiftx, shifty) for each channel] [nm]
        return [(0., 0.) for _ in range(self.nchannels)]

    #Shearing
    @metadatum(unit='nm/nm')
    def channel_shears(self) : #[(shearx, sheary) for each channel] []
        return [(0., 0.) for _ in range(self.nchannels)]

    #Rotations
    @metadatum(unit='°')
    def channel_angles(self) : #[theta for each channel] [°]
        return [0. for _ in range(self.nchannels)]

    #Zoom
    @metadatum(unit='')
    def channel_scales(self) : #[(scalex, scaley) for each channel] []
        return [(1., 1.) for _ in range(self.nchannels)]

    #Transform matrix
    @initproperty
    def transform_matrices(self) :
        matrices = [transform_matrix(shape, shiftx, shifty, shearx, sheary, angle, scalex, scaley)
                for shape, (shiftx, shifty), (shearx, sheary), angle, (scalex, scaley) in zip(self.channel_shapes, self.channel_shifts, self.channel_shears, self.channel_angles, self.channel_scales)]
        return [matrix if (matrix != np.eye(3)).any() else None for matrix in matrices]
    @metadatum(init=True)
    def do_transform(self) :
        for matrix in self.transform_matrices :
            if matrix is not None :
                return True
        return False



    @metadatum(init=True, unit='nm')
    def local_window(self) :
        return int(round(self.PSF_influence*2))



    #temporal_median
    tau_fact = 25 #Number of tau_ON to consider temporal window
    @metadatum(unit='frames')
    def temporal_window(self) :
        return int(np.ceil(self.tau_ON*self.tau_fact/self.exposure))



    #Detection parameters
    @defaultproperty
    def noise_factsmin(self) : #minimum noise factor for thresholding detected image
        return 1
    @defaultproperty
    def noise_factsstep(self) : #noise factor step for thresholding detected image
        return 0.5
    @defaultproperty
    def noise_factsmax(self) : #maximum noise factor for thresholding detected image
        return 2.001
    @metadatum(init=True)
    def noise_facts(self) : #can be set as a list if needed for different configs
        return np.arange(self.noise_factsmin, self.noise_factsmax, self.noise_factsstep)
    @metadatum(init=True)
    def area_min(self) :
        return 5
    @metadatum(init=True)
    def area_max(self) :
        radius = 2*int(np.ceil(2*self.PSF_sigma))+1
        return int(np.sum((ROI(field=radius, pixel=self.pixel, center=True).R <= radius)))


    #filters to apply
    @metadatum()
    def detfilters2apply(self) : #List of filter names to apply
        return ['close_borders']

    #filters to apply
    @metadatum()
    def locfilters2apply(self) : #List of filter names to apply
        return ['close_borders']



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-25
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : smlmLP
# Module        : block

"""
This function is a decorator to be used on block function, which allow to use config for default values.
"""



# %% Libraries
import functools
import time
import inspect
from smlmlp import columns



# %% Function
def block(timeit=True) :
    '''
    This function is a decorator to be used on block function, which allow to use config for default values.
    A decorated function can use a config object with config=config_object to defined default value of all the keyword only parameters.
    You can also use a Locs object via the locs=mylocsobject attribute for all parameters linked to localizations
    This decorator works for functions and generators.
    Computation time will be added for each call of the decorated function in block.times dictionary.
    
    Returns
    -------
    function : function
        Decorated function.

    Examples
    --------
    >>> from smlmlp import block
    ...
    >>> block()
    ... def myfunc() :
    ...     return long_process()
    ...
    >>> result = myfunc(config=config_object)
    '''

    def decorator(function) :
        name = function.__name__
        @functools.wraps(function)
        def wrapper(*args, config=None, locs=None, **kwargs) -> None :

            # Locs and config
            if locs is not None :
                if config is not None :
                    raise SyntaxError('Cannot call config and locs in block function')
                config = locs.config

            # Manages kwargs from config and locs
            kwargs = {key: value for key, value in kwargs.items() if value is not None}
            if config is not None :
                signature = inspect.signature(function)
                kw = {}
                for pname, param in signature.parameters.items() :
                    if (param.kind is inspect.Parameter.KEYWORD_ONLY or param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD) and pname in columns :
                        kw[pname] = getattr(locs.detections, pname)
                    elif (param.kind is inspect.Parameter.KEYWORD_ONLY or param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD) and hasattr(config, pname) :
                        kw[pname] = getattr(config, pname)
                kw.update(kwargs)
                kwargs = kw

            #Launch timed function
            tic = time.perf_counter()
            result = function(*args, **kwargs)
            toc = time.perf_counter()

            # Check if generator, if normal function just exit here
            if not inspect.isgenerator(result):
                if timeit :
                    if name in block.times :
                        block.times[name] += toc-tic
                    else :
                        block.times[name] = toc-tic
                return result
            
            # If is a generator
            def generator_wrapper():
                while True:
                    try:
                        tic = time.perf_counter()
                        value = next(result)
                        toc = time.perf_counter()
                        if timeit :
                            if name in block.times :
                                block.times[name] += toc-tic
                            else :
                                block.times[name] = toc-tic
                        yield value
                    except StopIteration:
                        break
            return generator_wrapper()

        return wrapper
    return decorator
block.times = {}



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



def metadatum(group, dtype=None, name=None, iterable=None):
    '''
    Decorator used in Config class when defining a metadatum property
    will store the datum in the metadata dictionnary
    '''

    def decorator(func) :
        datum = func.__name__ if name is None else name
        if group not in metadatum.groups : metadatum.groups[group] = []
        metadatum.groups[group].append(datum) # Groups of metadata

        #Getter
        def getter(self):
            _attribut = getattr(self, f'_{datum}', None)
            if _attribut is not None : return _attribut
            _attribute = func(self)
            return _attribute

        #Setter
        def setter(self, value):
            if iterable is not None :
                try :
                    if len(value) != iterable :
                        raise ValueError(f'Cannot have lenght {len(value)}')
                except TypeError :
                    value = [value for _ in range(iterable)]
                if dtype is not None :
                    value = [dtype(v) for v in value]
            else :
                if dtype is not None :
                    value = dtype(value)
            setattr(self, f'_{datum}', value)

        #Deleter
        def deleter(self):
            setattr(self, f'_{datum}', None)

        #Property
        return property(getter, setter, deleter)
    return decorator

#Adding attributes
metadatum.groups = {}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
import pandas as pd
import numpy as np



# %% Function
class DataFrame(pd.DataFrame) :
    '''
    Localization dimensional dataframe
    '''

    def __init__(self, locs) :
        if self.index_name is None : raise SyntaxError(f'DataFrame {self.__class__} should have an index name defined via @column decorator')
        self.locs = locs
        pd.DataFrame.__init__(self, index=np.sort(self.locs.detections[self.index_name].unique()))
        self.index.name = self.index_name
    
    # Attributes
    index_name = None # raise error if stays None
    locs = None # overriden in __init__

    def __setattr__(self, name, value):
        cls = type(self)
        attr = getattr(cls, name, None)

        # If class attribute is a descriptor with __set__
        if hasattr(attr, "__set__"):
            return attr.__set__(self, value)

        # Otherwise fallback to pandas
        return super().__setattr__(name, value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import block, Config
from arrlp import sortloop, get_xp, nb_threads
import numpy as np
import numba as nb
from numba import cuda as nb_cuda



# %% Function
@block()
def crop_individual_extract(channels, /, fr, x, y, ch=None, *, crop_size=11, channel_pixel=1., cuda=False, parallel=False) :
    '''
    This function creates the spatial local mean background.
    '''
    assert fr.min() >= 1, "Frame column starts at 1 by convention, please add 1 to your column frame before inserting it in this function"

    # Correct channels
    if ch is None :
        if len(channels) > 1 :
            raise SyntaxError("Cannot apply crop extracting on several channels without defining channel vector")
        ch = np.zeros_like(fr, dtype=np.uint8)

    # Get pixel
    fake_config = Config(nfiles=len(channels), pixel=channel_pixel, crop_size=crop_size)
    channel_pixel = fake_config.pixel
    crop_size = fake_config.crop_size

    # Sort
    xp = get_xp(cuda)
    argsort = xp.lexsort((x, y, fr, ch))
    fr = xp.asarray(fr[argsort]-1, dtype=xp.uint32)
    y = xp.asarray(y[argsort], dtype=xp.float32)
    x = xp.asarray(x[argsort], dtype=xp.float32)
    ch = xp.asarray(ch[argsort], dtype=xp.uint8)

    # Looping on channels
    crops, X0, Y0 = [], [], []
    for ch_ch, ch_fr, ch_y, ch_x in sortloop(ch, fr, y, x) :
        channel = channels[ch_ch[0]]
        pixel = channel_pixel[ch_ch[0]]
        size = crop_size[ch_ch[0]]
        width, height = int(np.ceil(size[0] / pixel[0])), int(np.ceil(size[1] / pixel[1]))
        n = len(ch_ch)
        crop = xp.empty(shape=(n, height, width), dtype=np.float32)
        x0_pix = xp.empty(shape=n, dtype=np.uint16)
        y0_pix = xp.empty(shape=n, dtype=np.uint16)

        if cuda :
            threads_per_block = (8, 8, 8)
            blocks_per_grid = (
                (n + threads_per_block[0] - 1) // threads_per_block[0],
                (height + threads_per_block[1] - 1) // threads_per_block[1],
                (width + threads_per_block[2] - 1) // threads_per_block[2],
            )
            crop_gpu[blocks_per_grid, threads_per_block](channel, crop, ch_fr, ch_x, ch_y, pixel[1], pixel[0], width, height, x0_pix, y0_pix)
        else :
            with nb_threads(parallel) :
                crop_cpu(channel, crop, ch_fr, ch_x, ch_y, pixel[1], pixel[0], width, height, x0_pix, y0_pix)
        crops.append(crop)
        X0.append(x0_pix)
        Y0.append(y0_pix)

    return crops, X0, Y0



@nb.njit(fastmath=True, cache=True, nogil=True, parallel=True)
def crop_cpu(channel, crop, F, X, Y, xpixel, ypixel, w, h, x0_pix, y0_pix) :
    YY, XX = channel[0].shape
    for i in nb.prange(len(crop)) :

        # bbox
        fr, x, y = F[i], X[i], Y[i]
        if w % 2 : #if odd
            xpix = int(x / xpixel + 0.5) # rounding
            x0 = xpix - w//2
        else : #if even
            xpix = int(x / xpixel) + 0.5 # rounding
            x0 = xpix - w//2 + 0.5
        if h % 2 : #if odd
            ypix = int(y / ypixel + 0.5) # rounding
            y0 = ypix - h//2
        else : #if even
            ypix = int(y / ypixel) + 0.5 # rounding
            y0 = ypix - h//2 + 0.5
        x0, y0 = int(x0), int(y0)

        # fill
        x0_pix[i] = x0
        y0_pix[i] = y0
        for yy in range(y0, y0 + h):
            for xx in range(x0, x0 + w):
                if 0 <= yy < YY and 0 <= xx < XX:
                    crop[i, yy-y0, xx-x0] = float(channel[fr, yy, xx])
                else :
                    crop[i, yy-y0, xx-x0] = 0.0



@nb_cuda.jit(fastmath=True)
def crop_gpu(channel, crop, F, X, Y, xpixel, ypixel, w, h, x0_pix, y0_pix) :
    i, dy, dx = nb_cuda.grid(3)
    n = crop.shape[0]
    if i < n and dy < h and dx < w:
        YY, XX = channel.shape[1], channel.shape[2]

        # bbox
        fr, x, y = F[i], X[i], Y[i]
        if w % 2 : #if odd
            xpix = int(x / xpixel + 0.5) # rounding
            x0 = xpix - w//2
        else : #if even
            xpix = int(x / xpixel) + 0.5 # rounding
            x0 = xpix - w//2 + 0.5
        if h % 2 : #if odd
            ypix = int(y / ypixel + 0.5) # rounding
            y0 = ypix - h//2
        else : #if even
            ypix = int(y / ypixel) + 0.5 # rounding
            y0 = ypix - h//2 + 0.5
        x0, y0 = int(x0), int(y0)

        # store origin (only once per crop)
        if dy == 0 and dx == 0:
            x0_pix[i] = x0
            y0_pix[i] = y0

        # global coordinates
        xx = x0 + dx
        yy = y0 + dy

        # bounds check
        if 0 <= yy < YY and 0 <= xx < XX:
            crop[i, dy, dx] = float(channel[fr, yy, xx])
        else:
            crop[i, dy, dx] = 0.0




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import MainDataFrame, column
import numpy as np



# %% Function
class detections(MainDataFrame) :
    '''
    Frames dataframe
    '''

    @column(headers=['detection', 'id'], save=True, index=True, agg='min')
    def det(self:np.uint64) :
        return None



    # --- Filters ---

    @column(headers=['filter'], save=False, index=False, agg='min')
    def keep(self:np.bool_) : # to be used with the self.filter function only
        return None



    # --- Detections ---

    @column(headers=['area [pix2]'], save=True, index=False, agg='mean')
    def area(self:np.uint8) :
        return None

    @column(headers=['x detection [nm]'], save=True, index=False, agg='mean')
    def xdet(self:np.float32) :
        return None

    @column(headers=['y detection [nm]'], save=True, index=False, agg='mean')
    def ydet(self:np.float32) :
        return None



    # --- Fits ---

    @column(headers=['x [nm]', 'xnm'], save=True, index=False, agg='mean')
    def x(self:np.float32) :
        return "xdet"

    @column(headers=['y [nm]', 'ynm'], save=True, index=False, agg='mean')
    def y(self:np.float32) :
        return "ydet"

    @column(headers=['z [nm]', 'znm'], save=True, index=False, agg='mean')
    def z(self:np.float32) :
        return "zdaisy"

    @column(headers=['chi2'], save=True, index=False, agg='max')
    def chi2(self:np.float32) :
        return None

    @column(headers=['iteration number'], save=True, index=False, agg='max')
    def n_iter(self:np.uint16) :
        return None

    @column(headers=['converged'], save=True, index=False, agg='min')
    def converged(self:np.bool_) :
        return None

    @column(headers=['fit amplitude [photon.pix-2]'], save=True, index=False, agg='mean')
    def amp(self:np.float32) :
        return None

    @column(headers=['fit offset [photon.pix-2]', 'offset [photon]'], save=True, index=False, agg='mean')
    def os(self:np.float32) :
        return None



    # --- Number of photons ---

    @column(headers=['signal [photon]', 'intensity [photon]', 'phot'], save=True, index=False, agg='sum')
    def signal(self:np.float32) :
        return "gaussiansignal"

    @column(headers=['gaussian signal [photon]'], save=False, index=False, agg='sum')
    def gaussiansignal(self:np.float32) :
        return self.amp / 2 / np.pi / (self.sigmax / self.config.xpixel) / (self.sigmay / self.config.ypixel)

    @column(headers=['bkgd [photon]', 'bg'], save=True, index=False, agg='sum')
    def bkgd(self:np.float32) :
        return "gaussianbkgd"

    @column(headers=['gaussian bkgd [photon]'], save=False, index=False, agg='sum')
    def gaussianbkgd(self:np.float32) :
        return self.os * self.config.cropsize**2

    @column(headers=['noise [photon]', 'bkgstd [photon]'], save=True, index=False, agg='mean')
    def noise(self:np.float32) :
        return np.sqrt(self.bkgd) #Poisson noise variance = expected



    # --- Gaussians size ---

    @column(headers=['Gaussian sigma [nm]', 'sigma [nm]'], save=True, index=False, agg='mean')
    def sigma(self:np.float32) :
        return np.sqrt(self.sigmax * self.sigmay)

    @column(headers=['Gaussian sigma x [nm]', 'sigma x [nm]', 'sigma_x [nm]', 'PSFxnm'], save=True, index=False, agg='mean')
    def sigmax(self:np.float32) :
        return "sigma"

    @column(headers=['Gaussian sigma y [nm]', 'sigma y [nm]', 'sigma_y [nm]', 'PSFynm'], save=True, index=False, agg='mean')
    def sigmay(self:np.float32) :
        return self.sigmax

    @column(headers=['sigma ratio', 'ellipticity'], save=False, index=False, agg='mean')
    def sigmaratio(self:np.float32) :
        return self.sigmax / self.sigmay

    @column(headers=['sigma angle [rad]', 'angle [rad]'], save=True, index=False, agg='mean')
    def sigmaangle(self:np.float32) :
        return None



    # --- CRLB ---

    @column(headers=['CRLB [nm]'], save=False, index=False, agg='mean')
    def crlb(self:np.float32) :
        return (self.crlb_xy**2 * self.crlb_z)**(1/3)

    @column(headers=['CRLB xy [nm]', 'uncertainty_xy'], save=False, index=False, agg='mean')
    def crlb_xy(self:np.float32) :
        # tau = 2*np.pi * (self.bkgd_std**2 + self.camera_compensation) * (self.sigmax * self.sigmay + self.pixel**2/12) / self.pixel**2 / self.signal
        # if self.fit_estimator == 'MLE' :
        #     return np.sqrt((self.camera_gain * self.sigmax*self.sigmay + (self.pixel**2/12)) / self.signal * (1 + 4*tau + np.sqrt(2*tau/(1+4*tau))))
        # if self.fit_estimator == 'LSE' :
        #     return np.sqrt((self.camera_gain * self.sigmax*self.sigmay + (self.pixel**2/12)) / self.signal * (4*tau + 16/9))
        # return None
        return None

    @column(headers=['CRLB z [nm]', 'uncertainty_z'], save=False, index=False, agg='mean')
    def crlb_z(self:np.float32) :
        return "crlb_xy"






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import block
from stacklp import temporal_median
import numpy as np



# %% Function
@block()
def bkgd_temporal_median(channels, /, median_window=25, bkgds=None, noise_corrections=None, *, exposure=1., cuda=False, parallel=False) :
    '''
    This function creates the temporal local median background.
    '''

    # Get temporal window in frames
    median_window = int(round(median_window / exposure))

    # Correct bkgd length for end of acquisition
    if bkgds is not None and len(bkgds[0]) > len(channels[0]):
        bkgds = [bkgd[:len(channel)] for channel, bkgd in zip(channels, bkgds)]
    
    # Noise corrections
    if noise_corrections is None :
        noise_corrections = [np.float32(1.) for _ in range(len(channels))]

    new_bkgds = []
    for i in range(len(channels)) :
        bkgd = None if bkgds is None else bkgds[i]
        channel = channels[i]
        new_bkgd = temporal_median(channel, median_window, out=bkgd, cuda=cuda, parallel=parallel)
        noise_corrections[i] *= np.float32(1.)
        new_bkgds.append(new_bkgd)

    return new_bkgds, noise_corrections


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries




# %% Function
def temp(**kwargs) :
    '''
    TODO
    '''

    print('Hello world!')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



import csv



def save_df(df, path, head2save=None) :

    #Add quotes
    columns = list(df.columns)
    for pos,col in enumerate(columns) :
        columns[pos] = '"' + col + '"'
    df.columns = columns

    #Save df
    if head2save is not None :
        columns = [col for col in df.columns if col[1:-1] in head2save]
    df.to_csv(path, columns=columns, quoting = csv.QUOTE_NONE, float_format="%.3f")

    #Substract quotes
    columns = list(df.columns)
    for pos,col in enumerate(columns) :
        columns[pos] = col[1:-1]
    df.columns = columns




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import block, signal_spatial_filter, signal_temporal_filter
from arrlp import gc



# %% Function
@block(timeit=False)
def signal_combination(channels, /, spatial_kernel=None, temporal_kernel=None, signals=None, bkgds=None, noise_corrections=None, *, do_spatial_filter=True, do_temporal_filter=False, cuda=False, parallel=False) :
    '''
    This function creates the spatial local mean background.
    '''

    # init
    buffers = None

    if do_spatial_filter :
        assert spatial_kernel is not None
        gc()
        signals, noise_corrections = signal_spatial_filter(channels, spatial_kernel, signals=signals, noise_corrections=noise_corrections, bkgds=bkgds, cuda=cuda, parallel=parallel)
        if buffers is None :
            buffers = [signal.copy() for signal in signals]
        else :
            for i in range(len(channels)) :
                buffers[i][:] = signals[i]
        channels = buffers
        bkgds = None
    if do_temporal_filter :
        assert temporal_kernel is not None
        gc()
        signals, noise_corrections = signal_temporal_filter(channels, temporal_kernel, signals=signals, noise_corrections=noise_corrections, bkgds=bkgds, cuda=cuda, parallel=parallel)
        if buffers is None :
            buffers = [signal.copy() for signal in signals]
        else :
            for i in range(len(channels)) :
                buffers[i][:] = signals[i]
        channels = buffers
        bkgds = None
    gc()

    return signals, noise_corrections


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-08-28
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : smlmLP

"""
A python library for Single Molecule Localization Microscopy.
"""



# %% Source import
sources = {
'Config': 'smlmlp.modules.Config_LP.Config',
'Locs': 'smlmlp.modules.Locs_LP.Locs',
'block': 'smlmlp.modules.block_LP.block',
'columns': 'smlmlp.modules.columns_LP.columns',
'computer': 'smlmlp.modules.computer_LP.computer',
'dataframes': 'smlmlp.modules.dataframes_LP.dataframes',
'Camera': 'smlmlp.modules.Config_LP._functions.Camera',
'metadatum': 'smlmlp.modules.Config_LP._functions.metadatum',
'Channel': 'smlmlp.modules.Config_LP._functions.Channel',
'save_df': 'smlmlp.modules.Locs_LP._functions.save_df',
'open_df': 'smlmlp.modules.Locs_LP._functions.open_df',
'bkgd_spatial_mean': 'smlmlp.modules.block_LP._functions.background.bkgd_spatial_mean',
'bkgd_spatial_opening': 'smlmlp.modules.block_LP._functions.background.bkgd_spatial_opening',
'bkgd_combination': 'smlmlp.modules.block_LP._functions.background.bkgd_combination',
'bkgd_temporal_median': 'smlmlp.modules.block_LP._functions.background.bkgd_temporal_median',
'crop_individual_extract': 'smlmlp.modules.block_LP._functions.crop.crop_individual_extract',
'load_data': 'smlmlp.modules.block_LP._functions.loading.load_data',
'signal_spatial_filter': 'smlmlp.modules.block_LP._functions.signal.signal_spatial_filter',
'signal_combination': 'smlmlp.modules.block_LP._functions.signal.signal_combination',
'signal_temporal_filter': 'smlmlp.modules.block_LP._functions.signal.signal_temporal_filter',
'blink_spatial_psf': 'smlmlp.modules.block_LP._functions.blink.blink_spatial_psf',
'blink_temporal_on': 'smlmlp.modules.block_LP._functions.blink.blink_temporal_on',
'detect_spatial_maxima': 'smlmlp.modules.block_LP._functions.detection.detect_spatial_maxima',
'detect_gain': 'smlmlp.modules.block_LP._functions.detection.detect_gain',
'detect_snr': 'smlmlp.modules.block_LP._functions.detection.detect_snr',
'column': 'smlmlp.modules.columns_LP._functions.column',
'LocsDataFrame': 'smlmlp.modules.dataframes_LP._functions.LocsDataFrame',
'MainDataFrame': 'smlmlp.modules.dataframes_LP._functions.MainDataFrame',
'DetsDataFrame': 'smlmlp.modules.dataframes_LP._functions.DetsDataFrame',
'DataFrame': 'smlmlp.modules.dataframes_LP._functions.DataFrame'
}



# %% Lazy imports
from corelp import getmodule
__getattr__, __all__ = getmodule(sources)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import block, Config, bkgd_spatial_opening, bkgd_temporal_median, bkgd_spatial_mean
from arrlp import gc, get_xp



# %% Function
@block(timeit=False)
def bkgd_combination(channels, /, bkgds=None, noise_corrections=None, *, do_spatial_opening=False, opening_radius=5, do_temporal_median=True, median_window=25, do_spatial_mean=True, mean_radius=7, channel_pixel=1., exposure=1., cuda=False, parallel=False) :
    '''
    This function creates the spatial local mean background.
    '''

    # init
    raws = channels
    buffers = None
    xp = get_xp(cuda)

    if do_spatial_opening :
        gc()
        bkgds, noise_corrections = bkgd_spatial_opening(channels, opening_radius=opening_radius, noise_corrections=noise_corrections, bkgds=bkgds, cuda=cuda, parallel=parallel, channel_pixel=channel_pixel)
        if buffers is None :
            channels = [channel - bkgd for channel, bkgd in zip(channels, bkgds)]
            buffers = channels
        else :
            for i in range(len(channels)) :
                xp.subtract(channels[i], bkgds[i], buffers[i])
    if do_temporal_median :
        gc()
        bkgds, noise_corrections = bkgd_temporal_median(channels, median_window=median_window, noise_corrections=noise_corrections, bkgds=bkgds, cuda=cuda, parallel=parallel, exposure=exposure)
        if buffers is None :
            channels = [channel - bkgd for channel, bkgd in zip(channels, bkgds)]
            buffers = channels
        else :
            for i in range(len(channels)) :
                xp.subtract(channels[i], bkgds[i], buffers[i])
    if do_spatial_mean :
        gc()
        bkgds, noise_corrections = bkgd_spatial_mean(channels, mean_radius=mean_radius, noise_corrections=noise_corrections, bkgds=bkgds, cuda=cuda, parallel=parallel, channel_pixel=channel_pixel)
        if buffers is None :
            channels = [channel - bkgd for channel, bkgd in zip(channels, bkgds)]
            buffers = channels
        else :
            for i in range(len(channels)) :
                xp.subtract(channels[i], bkgds[i], buffers[i])
    gc()
    if buffers is None :
        if bkgds is None :
            bkgds = [xp.zeros_like(channel) for channel in channels]
        else :
            for bkgd in bkgds :
                bkgd[:] = 0
    else :
        for i in range(len(channels)) :
            xp.subtract(raws[i], channels[i], bkgds[i])

    return bkgds, noise_corrections


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import block
from arrlp import get_xp, img_correlate
import numpy as np



# %% Function
@block()
def signal_spatial_filter(channels, /, spatial_kernel, signals=None, bkgds=None, noise_corrections=None, *, cuda=False, parallel=False) :
    '''
    This function applyies a spatial filter to enhance signal.
    '''

    # Correct signal length for end of acquisition
    if signals is not None and len(signals[0]) > len(channels[0]):
        signals = [signal[:len(channel)] for channel, signal in zip(channels, signals)]
    
    # Noise corrections
    if noise_corrections is None :
        noise_corrections = [np.float32(1.) for _ in range(len(channels))]

    new_signals = []
    for i in range(len(channels)) :
        bkgd = None if bkgds is None else bkgds[i]
        signal = None if signals is None else signals[i]
        channel = channels[i]
        if bkgd is not None :
            channel = channel - bkgd
        if signal is channel : signal = None
        kernel = spatial_kernel[i]
        new_signal = img_correlate(channel, kernel=kernel, out=signal, cuda=cuda, parallel=parallel, stacks=True)
        noise_corrections[i] *= np.sqrt(np.sum(kernel**2))
        new_signals.append(new_signal)

    return new_signals, noise_corrections


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import block, Config
from arrlp import img_gaussianfilter, kernel
import numpy as np



# %% Function
@block()
def bkgd_spatial_mean(channels, /, mean_radius=7, bkgds=None, noise_corrections=None, *, channel_pixel=1., cuda=False, parallel=False) :
    '''
    This function creates the spatial local mean background.
    '''

    # Get pixel
    channel_pixel = Config(nfiles=len(channels), pixel=channel_pixel).pixel

    # Correct bkgd length for end of acquisition
    if bkgds is not None and len(bkgds[0]) > len(channels[0]):
        bkgds = [bkgd[:len(channel)] for channel, bkgd in zip(channels, bkgds)]
    
    # Noise corrections
    if noise_corrections is None :
        noise_corrections = [np.float32(1.) for _ in range(len(channels))]

    new_bkgds = []
    for i in range(len(channels)) :
        bkgd = None if bkgds is None else bkgds[i]
        channel = channels[i]
        new_bkgd = img_gaussianfilter(channel, sigma=mean_radius/2, pixel=channel_pixel[i], out=bkgd, cuda=cuda, parallel=parallel, stacks=True)
        k1 = -kernel(ndims=1, pixel=channel_pixel[i][0], sigma=mean_radius/2)
        k2 = -kernel(ndims=1, pixel=channel_pixel[i][1], sigma=mean_radius/2)
        k1[int(len(k1)//2)] += 1.
        k2[int(len(k2)//2)] += 1.
        noise_corrections[i] *= np.sqrt(np.sum(k1**2)) * np.sqrt(np.sum(k2**2))
        new_bkgds.append(new_bkgd)

    return new_bkgds, noise_corrections


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import DataFrame, column
import numpy as np



# %% Function
class dyes(DataFrame) :
    '''
    Dyes dataframe
    '''

    @column(headers=['dye'], save=True, index=True, agg='min')
    def dye(self:np.uint8) :
        return None


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import block
from contextlib import ExitStack
from arrlp import gc
import tifffile as tiff
from stacklp import shapetif
import numpy as np
from concurrent.futures import ThreadPoolExecutor



# %% Function
@block()
def load_data(*tif_paths, chunk=None, pad=0, bbox=None, memmap=True, flip=None, iterator=range) :
    '''
    This generator loads SMLM raw data in chunks from various tif files.
    '''

    with ExitStack() as stack :
        block.times = {} # Reset timings of blocks
        tifs = [stack.enter_context(tiff.TiffFile(file)) for file in tif_paths]
        shapes = [shapetif(tif) for tif in tifs]
        nfiles = len(tifs)
        if nfiles < 1 : raise SyntaxError('Must define at least one tiff file to load')
        if bbox is None : bbox = [[(0, 0, shape[2], shape[1]),] for shape in shapes]
        nchannels = [len(box) for box in bbox]
        if len(nchannels) != nfiles : raise ValueError('Did not give the same amount of bbox as files')

        # Check number of frames
        nframes = None
        for shape in shapes :
            if nframes is None : nframes = shape[0]
            if len(shape) != 3 :
                raise ValueError(f'Tiff files for SMLM data should have 3 dimensions (time, y, x), not {shape}')
            if shape[0] != nframes :
                raise ValueError('All tiff files do not have the same number of frames which is not possible for a single SMLM acquisition')

        # Correct parameters
        if chunk is None : chunk = nframes
        if nframes < chunk :
            chunk = nframes
        if nframes <= pad :
            pad = nframes - 1
        nloops = int(np.ceil(nframes / chunk))

        # Try memmap
        if memmap :
            mmaps = []
            for tif_path in tif_paths :
                try :
                    mmap = tiff.memmap(tif_path)
                except ValueError :
                    mmap = None
                mmaps.append(mmap)
        else :
            mmaps = [None for _ in range(nfiles)]

        # Allocating memory
        shapes = [shape[1:3] for shape in shapes] # Remove number of frames
        dtypes = [tif.pages.get(0).dtype for tif in tifs] # Get dtype of each file
        loads = [np.empty(shape=(chunk + 2 * pad, *shape), dtype=dtype) if mmap is None else None for shape, dtype, mmap in zip(shapes, dtypes, mmaps)]

        # loops
        for loop in iterator(nloops) :
            gc()  # Force garbage collection

            #             <-- data flux <--          
            # | 00pad01 |     0chunk1     | 10pad11 | BEFORE
            # |                loaded               |
            # |unk1     | 10pad11 |    new|array    | AFTER

            # Calculate borders
            chunk0, chunk1 = loop * chunk, (loop+1) * chunk -1
            pad00, pad01 = chunk0 - pad, chunk0 - 1
            pad10, pad11 = chunk1 + 1, chunk1 + pad
            borders = dict(chunk0=chunk0, chunk1=chunk1, pad00=pad00, pad01=pad01, pad10=pad10, pad11=pad11)



            # Memmaps chunks
            mmaps_chunks = [mmap[max(0, pad00) : min(pad11 + 1, nframes)] if mmap is not None else None for mmap in mmaps]


            # Load
            with ThreadPoolExecutor(max_workers=nfiles) as pool:
                futures = [
                    pool.submit(_load_one_tif, tif, load, loop, chunk, pad, nframes, chunk1, pad10, pad11)
                    for tif, load, mmap in zip(tifs, loads, mmaps) if mmap is None
                    ]

                for f in futures:
                    f.result()

            # Slicing if end of acquisition
            if pad11 + 1 > nframes :
                loads = [load[:nframes - pad11 - 1] if load is not None else None for load in loads]



            # --- Dividing loads arrays into channels ---



            # Make channel views
            channels = []
            count = 0
            for load, mmap, box in zip(loads, mmaps_chunks, bbox) :
                for bb in box :
                    # bb = (x0, y0, x1, y1) slicing --> [y0:y1, x0:x1]
                    x0, y0, x1, y1 = bb
                    channel = load[:, y0:y1, x0:x1] if mmap is None else mmap[:, y0:y1, x0:x1]
                    if flip is not None :
                        if flip[count][0] : channel = channel[:, ::-1, :]
                        if flip[count][1] : channel = channel[:, :, ::-1]
                    channels.append(channel)
                    count += 1



            # Return the generator value
            yield channels, borders



def _load_one_tif(tif, load, loop, chunk, pad, nframes, chunk1, pad10, pad11) :

    # Defining views
    array_pad0 = load[:pad]
    array_chunk = load[-pad-chunk:len(load)-pad]
    array_pad1 = load[len(load)-pad:]

    # First loop scenario
    if pad > 0 and loop == 0 :
        tif.asarray(key=slice(0, pad, 1), out=array_pad1)

    # Transfering already loaded data
    if pad > 0 :
        np.copyto(array_pad0, array_chunk[-pad:], casting='no')
        np.copyto(array_chunk[:pad,:,:], array_pad1, casting='no')

    # Loading chunk data
    pos0 = min(chunk1 + pad - chunk + 1, nframes)
    pos1 = min(chunk1 + 1, nframes)
    delta = pos1 - pos0
    if pos0 < nframes :
        tif.asarray(key=slice(pos0, pos1, 1), out=array_chunk[pad:pad+delta,:,:])

    # Loading pad1 data
    if pad > 0 :
        pos0 = min(pad10, nframes)
        pos1 = min(pad11 + 1, nframes)
        delta = pos1 - pos0
        if pos0 < nframes :
            tif.asarray(key=slice(pos0, pos1, 1), out=array_pad1[:delta,:,:])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import dataframes
DetsDataFrame = dataframes['detections']


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from corelp import prop



# %% Function
class Camera :
    '''
    Defines camera instance
    '''

    metadata = [ # (metadata, group)
        ("nchannels", "Cameras"),
        ("npixels", "Cameras"),
        ("bits", "Cameras"),
        ("pixel", "Cameras"),
        ("gain_constructor", "Cameras"),
        ("gain_experimental", "Backgrounds"),
        ("QE", "Cameras"),
        ]
    properties = ['camera_index', 'gain']



    def __init__(self, config) :
        self.config = config



    # Channels
    @property
    def camera_index(self) :
        for i in range(self.config.nfiles) :
            if self.config.cameras[i] is self :
                return i

    @property
    def channels(self) :
        if not hasattr(self, '_channels') : self.nchannels = 1
        return self._channels
    @property
    def nchannels(self) :
        return len(self.channels)
    @nchannels.setter
    def nchannels(self, value) :
        from smlmlp import Channel
        self._channels = [Channel(self) for _ in range(int(value))]
    @property
    def _nchannels(self) :
        channels = getattr(self, '_channels', None)
        return None if channels is None else len(channels)



    # Pixels

    @prop(iterable=2, dtype=int)
    def npixels(self) : # (y, x)
        return 2304

    @prop(iterable=2, dtype=float)
    def pixel(self) : # (y, x) [nm]
        return 100.



    # Data

    @prop(dtype=int)
    def bits(self) :
        return 16

    @property
    def frame_bytes(self) : # gigabytes/frame
        return self.npixels[0] * self.npixels[1] * self.bits / 8 / 1024**3
    


    # Photons counting
    @property
    def gain(self) :
        return self.gain_experimental if self.gain_experimental is not None else self.gain_constructor

    @prop(dtype=float)
    def gain_constructor(self) : # e-/ADU (Analog to Digital Unit)
        return 0.25

    @prop(dtype=float)
    def gain_experimental(self) : # e-/ADU (Analog to Digital Unit)
        return None

    @prop(dtype=float)
    def QE(self) : # Quantum Efficiency
        0.8
    


    @property
    def FOV_max(self) : # (y, y) [µm]
        ny, nx = self.npixels
        match self.nchannels :
            case 1 : # Full frame
                pass
            case 2 : # Divide along x direction (fast sCMOS axis, we want each channel to have the same time stamps per line)
                nx = nx - nx%2 # Make it even
                nx = nx // 2
            case 3 : # Divide along x direction, same reason
                nx = nx - nx%3 # Make it 3 multiple
                nx = nx // 3
            case 4 : # 4 quadrants
                nx = nx - nx%2 # Make it even
                ny = ny - ny%2 # Make it even
                nx, ny = nx // 2, ny // 2
            case _ : raise ValueError('Dividing a camera image into more than 4 channels is not supported.')
        return int(ny) * self.pixel[0] * 1e-3, int(nx) * self.pixel[1] * 1e-3

    def FOV2bbox(self, FOV) : # (x0, y0, x1, y1)
        ny, nx = self.npixels # total number of pixels in frames
        sy, sx = int(FOV[0] * 1e3 / self.pixel[0]), int(FOV[1] * 1e3 / self.pixel[1]) # shape of each channel
        match self.nchannels:
            case 1: # Center a single crop
                x0 = (nx - sx) // 2
                y0 = (ny - sy) // 2
                return [(x0, y0, x0 + sx, y0 + sy)]
            case 2: # Two horizontal channels (split x into 2)
                total_w = 2 * sx
                x_start = (nx - total_w) // 2
                y0 = (ny - sy) // 2
                return [(x_start + i * sx, y0, x_start + (i + 1) * sx, y0 + sy) for i in range(2)]
            case 3: # Three horizontal channels (split x into 3)
                total_w = 3 * sx
                x_start = (nx - total_w) // 2
                y0 = (ny - sy) // 2
                return [ (x_start + i * sx, y0, x_start + (i + 1) * sx, y0 + sy) for i in range(3)]
            case 4: # 2x2 grid (split x and y)
                total_w = 2 * sx
                total_h = 2 * sy
                x_start = (nx - total_w) // 2
                y_start = (ny - total_h) // 2
                boxes = []
                for j in range(2):      # y direction
                    for i in range(2):  # x direction
                        x0 = x_start + i * sx
                        y0 = y_start + j * sy
                        boxes.append((x0, y0, x0 + sx, y0 + sy))
                return boxes
            case _: raise ValueError("Dividing a camera image into more than 4 channels is not supported.")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-20
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : smlmLP
# Module        : Locs

"""
This file allows to test Locs

Locs : This class define objects corresponding to localizations sets for one experiment.
"""



# %% Libraries
from corelp import debug
import numpy as np
from smlmlp import Locs
debug_folder = debug(__file__)



# %% Function test
def test_function() :
    '''
    Test Locs function
    '''
    
    # Init
    locs = Locs()

    # Simulate random data
    argunsort = np.arange(10000)
    np.random.shuffle(argunsort)
    array = np.hstack([np.full(100, i) for i in range(100)]) # Stack of constants to apply to xdet and fr (should be paired)
    xdet = array.astype(np.float32)[argunsort]
    fr = array.astype(np.uint32)[argunsort]

    # Set data in locs object
    locs.detections.fr = fr
    locs.detections.xdet = xdet
    assert (locs.detections.xdet == xdet).all()
    assert (locs.detections.fr == fr).all()

    # Looking at frames dataframe with merging and spreads
    xdet_frames = locs.frames.xdet # merging
    assert (locs.frames.fr == xdet_frames).all()
    locs.frames.dx = xdet_frames # spreading
    assert (locs.detections.dx == xdet).all()

    # Saving and loading
    path = debug_folder / 'locs'
    locs.save(path)
    locs2 = Locs(path)



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from corelp import prop
from smlmlp import Camera
import numpy as np
from arrlp import coordinates




# %% Function
class Channel :
    '''
    Defines channel instance
    '''

    metadata = [ # (metadata, group)
        ("flip", "Cameras"),
        ("psf_sigx", "Blinks"),
        ("psf_sigy", "Blinks"),
        ("psf_theta", "Blinks"),
        ("psf_tx", "Data"),
        ("psf_ty", "Data"),
        ("psf_coeffs", "Data"),
        ]
    properties = ["psf_sigma", "psf_radius", "psf_diameter", "psf_fwhm", "spatial_kernel", "spatial_kernel_shape"]



    def __init__(self, camera) :
        self.camera = camera



    # Bounding box

    @prop(iterable=2, dtype=bool)
    def flip(self) :
        return False



    # PSF

    @prop()
    def psf_tx(self) : # spline
        return None
    @prop()
    def psf_ty(self) : # spline
        return None
    @prop()
    def psf_coeffs(self) : # spline
        return None

    @prop()
    def psf_sigx(self) : # [nm]
        return 0.21 * 670 / 1.5
    @prop()
    def psf_sigy(self) : # [nm]
        return 0.21 * 670 / 1.5
    @prop()
    def psf_theta(self) : # [°]
        return 0.

    @property
    def psf_sigma(self) : # [nm]
        return np.sqrt(self.psf_sigx * self.psf_sigy)
    @psf_sigma.setter
    def psf_sigma(self, value) :
        self.psf_sigx, self.psf_sigy = value, value
    @property
    def psf_wl_na(self) : # [nm]
        return self.sigma / 0.21
    @psf_wl_na.setter
    def psf_wl_na(self, value) :
        self.psf_sigma = 0.21 * value
    @property
    def psf_radius(self) : # nm
        return 0.61 * self.psf_wl_na
    @psf_radius.setter
    def psf_radius(self, value) :
        self.psf_wl_na = value / 0.61
    @property
    def psf_diameter(self) : # nm
        return 1.22 * self.psf_wl_na
    @psf_diameter.setter
    def psf_diameter(self, value) :
        self.psf_wl_na = value / 1.22
    @property
    def psf_fwhm(self) : # nm
        return 0.51 * self.psf_wl_na
    @psf_fwhm.setter
    def psf_fwhm(self, value) :
        self.psf_wl_na = value / 0.51




    # Spatial kernel

    @property
    def spatial_subtract_factor(self) :
        return self.camera.config.spatial_subtract_factor

    @property
    def spatial_kernel_shape(self) :
        sigma_pix = max(self.psf_sigx, self.psf_sigy) / min(self.pixel)
        if self.spatial_subtract_factor > 1 : sigma_pix *= self.spatial_subtract_factor
        sigma_pix = int(np.ceil(sigma_pix))
        return (sigma_pix * 6 + 1, sigma_pix * 6 + 1)

    @property
    def psf_kernel(self) :
        Y, X = coordinates(shape=self.spatial_kernel_shape, pixel=self.pixel, grid=False)
        if self.psf_tx is not None and self.psf_ty is not None and self.psf_coeffs is not None :
            from funclp import Spline2D
            tx, ty = np.asarray(self.psf_tx), np.asarray(self.psf_ty)
            coeffs = np.asarray(self.psf_coeffs).reshape(ty.size-4, tx.size-4) # 4 = k + 1 (order + 1)
            spline = Spline2D(tx=tx, ty=ty, coeffs=coeffs)
            k = spline(X, Y)
        else :
            from funclp import Gaussian2D
            gaussian = Gaussian2D(sigx=self.psf_sigx, sigy=self.psf_sigy, theta=self.psf_theta, pixx=self.pixel[1], pixy=self.pixel[0])
            k = gaussian(X, Y)
        k /= k.sum()
        return k.astype(np.float32)

    @property
    def spatial_subtract_kernel(self) :
        if self.spatial_subtract_factor <= 1 : return None
        pixel = self.pixel[0] / self.spatial_subtract_factor, self.pixel[1] / self.spatial_subtract_factor
        Y, X = coordinates(shape=self.spatial_kernel_shape, pixel=pixel, grid=False)
        if self.psf_tx is not None and self.psf_ty is not None and self.psf_coeffs is not None :
            from funclp import Spline2D
            tx, ty = np.asarray(self.psf_tx), np.asarray(self.psf_ty)
            coeffs = np.asarray(self.psf_coeffs).reshape(ty.size-4, tx.size-4) # 4 = k + 1 (order + 1)
            spline = Spline2D(tx=tx, ty=ty, coeffs=coeffs)
            k = spline(X, Y)
        else :
            from funclp import Gaussian2D
            gaussian = Gaussian2D(sigx=self.psf_sigx, sigy=self.psf_sigy, theta=self.psf_theta, pixx=self.pixel[1], pixy=self.pixel[0])
            k = gaussian(X, Y)
        k /= k.sum()
        return k.astype(np.float32)

    @prop(cache=True)
    def spatial_kernel(self) :
        if self.spatial_subtract_factor <= 1 :
            return self.psf_kernel
        k = self.psf_kernel - self.spatial_subtract_kernel
        return k - k.mean()



    # Crop

    @prop(iterable=2, dtype=int)
    def crop_size(self) :
        h = self.psf_y / self.pixel[0] * 12
        w = self.psf_x / self.pixel[1] * 12
        return int(2*(h//2)+1), int(2*(w//2)+1)



# Adding Camera metadata
for data, _ in Camera.metadata :
    @property
    def camera_property(self, data=data) :
        return getattr(self.camera, data)
    setattr(Channel, data, camera_property)
for data in Camera.properties :
    @property
    def camera_property(self, data=data) :
        return getattr(self.camera, data)
    setattr(Channel, data, camera_property)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-23
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : smlmLP
# Module        : columns

"""
This is a dict of all the columns that are created.
"""



class Columns(dict) :
    def __init__(self, *args, **kwargs) :
        super().__init__(*args, **kwargs)
        self.headers = {}
        self.head2save = []
columns = Columns()


# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import block, Config
from arrlp import get_xp, nb_threads
import numba as nb
from numba import cuda as nb_cuda 
import numpy as np
import math



# %% Function
@block()
def detect_spatial_maxima(snrs, /, snr_thresh, spatial_kernel, *, pad=0, f0=0, channel_pixel=1., cuda=False, parallel=False) :
    '''
    This function finds local maxima in thresholded areas.
    '''

    # xp
    xp = get_xp(cuda)

    # footprint
    k_fp = math.exp(-(0.47/0.21)**2 / 2) # Sparrow limit: 0.47 wl / NA, Gaussian 0.21 wl / NA
    footprints = [kernel > kernel.max() * k_fp for kernel in spatial_kernel]
    
    # Get pixel
    pixels = Config(nfiles=len(snrs), pixel=channel_pixel).pixel

    F, X, Y, C = [], [], [], []
    for pos, (snr, footprint, pixel) in enumerate(zip(snrs, footprints, pixels)) :
        snr = xp.asarray(snr)
        footprint = xp.asarray(footprint)

        # GPU detection
        if cuda :
            max_points = int(snr.size / xp.sum(footprint)) # upper bound
            fr_out = xp.empty(max_points, dtype=xp.int32)
            y_out = xp.empty(max_points, dtype=xp.float32)
            x_out = xp.empty(max_points, dtype=xp.float32)
            counter = xp.zeros(1, dtype=xp.int32)
            shape = snr.shape
            threads_per_block = (2, 16, 16)
            blocks_per_grid = (
                (shape[0] + threads_per_block[0] - 1) // threads_per_block[0],
                (shape[1]  + threads_per_block[1] - 1) // threads_per_block[1],
                (shape[2]  + threads_per_block[2] - 1) // threads_per_block[2],
            )
            det_gpu[blocks_per_grid, threads_per_block](snr, snr_thresh, footprint, fr_out, y_out, x_out, counter)
            n = xp.asnumpy(counter)[0]
            fr = xp.asnumpy(fr_out[:n])
            y = xp.asnumpy(y_out[:n])
            x = xp.asnumpy(x_out[:n])

        # CPU detection
        else :
            mask = xp.zeros_like(snr, dtype=xp.bool_)
            with nb_threads(parallel) :
                maxi_cpu(mask, snr, snr_thresh, footprint)
            fr, y_int, x_int = xp.nonzero(mask)
            y, x = xp.empty_like(y_int, dtype=xp.float32), xp.empty_like(x_int, dtype=xp.float32)

            if len(f) :
                with nb_threads(parallel) :
                    com_cpu(snr, fr, y_int, x_int, y, x)

        # Append
        fr += f0 + 1
        y *= pixel[0]
        x *= pixel[1]
        c = np.full_like(fr, fill_value=pos, dtype=np.uint8)
        argsort = np.lexsort((x, y, fr))
        F.append(fr[argsort])
        Y.append(y[argsort])
        X.append(x[argsort])
        C.append(c)

    return np.hstack(F), np.hstack(X), np.hstack(Y), np.hstack(C)



@nb.njit(parallel=True, nogil=True, cache=True, fastmath=True)
def maxi_cpu(mask, snr, snr_thresh, footprint):
    F, Y, X = snr.shape
    YY, XX = footprint.shape
    cy, cx = YY // 2, XX // 2

    for fr in nb.prange(F):
        frame = snr[fr]
        m = mask[fr]

        for y in range(Y):
            for x in range(X):

                val = frame[y, x]
                if val < snr_thresh:
                    continue

                is_max = True

                for dy in range(YY):
                    for dx in range(XX):

                        if not footprint[dy, dx]:
                            continue

                        ny = y + dy - cy
                        nx = x + dx - cx

                        if ny < 0 or ny >= Y or nx < 0 or nx >= X:
                            continue

                        if frame[ny, nx] > val:
                            is_max = False
                            break

                    if not is_max:
                        break

                if is_max:
                    m[y, x] = True



@nb.njit(parallel=True, nogil=True, cache=True, fastmath=True)
def com_cpu(snr, fr_idx, y_idx, x_idx, y_out, x_out):
    n = len(f_idx)

    F, Y, X = snr.shape

    for i in nb.prange(n):
        fr = fr_idx[i]
        y = y_idx[i]
        x = x_idx[i]

        xnum = 0.0
        ynum = 0.0
        denom = 0.0

        for dy in range(-1, 2):
            for dx in range(-1, 2):

                ny = y + dy
                nx = x + dx

                if ny < 0 or ny >= Y or nx < 0 or nx >= X:
                    continue

                val = snr[fr, ny, nx]

                xnum += nx * val
                ynum += ny * val
                denom += val

        if denom > 0:
            x_out[i] = xnum / denom
            y_out[i] = ynum / denom
        else :
            x_out[i] = x
            y_out[i] = y



@nb_cuda.jit
def det_gpu(snr, snr_thresh, footprint, fr_out, y_out, x_out, counter):
    fr, y, x = nb_cuda.grid(3)

    F, Y, X = snr.shape
    YY, XX = footprint.shape

    if fr >= F or y >= Y or x >= X:
        return

    val = snr[fr, y, x]
    if val < snr_thresh:
        return

    cy = YY // 2
    cx = XX // 2

    # ---- local max ----
    is_max = True

    for dy in range(YY):
        for dx in range(XX):

            if not footprint[dy, dx]:
                continue

            ny = y + dy - cy
            nx = x + dx - cx

            if ny < 0 or ny >= Y or nx < 0 or nx >= X:
                continue

            if snr[fr, ny, nx] > val:
                is_max = False
                break

        if not is_max:
            break

    if not is_max:
        return

    # ---- center of mass ----
    xnum = 0.0
    ynum = 0.0
    denom = 0.0

    for dy in range(-1, 2):
        for dx in range(-1, 2):

            ny = y + dy
            nx = x + dx

            if ny < 0 or ny >= Y or nx < 0 or nx >= X:
                continue

            v = snr[fr, ny, nx]
            xnum += nx * v
            ynum += ny * v
            denom += v

    if denom > 0:
        xf = xnum / denom
        yf = ynum / denom
    else:
        xf = x
        yf = y

    # ---- atomic append ----
    idx = nb_cuda.atomic.add(counter, 0, 1)

    fr_out[idx] = fr
    y_out[idx] = yf
    x_out[idx] = xf


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import DataFrame, column
import numpy as np



# %% Function
class frames(DataFrame) :
    '''
    Frames dataframe
    '''

    @column(headers=['frame', 'frames'], save=True, index=True, agg='min')
    def fr(self:np.uint32) :
        return None



    # --- Drifts ---

    @column(headers=['nodrift [nm]'], save=False, index=False, agg='mean')
    def nodrift(self:np.float32) :
        return np.zeros(self.nframes, dtype=np.float32)

    @column(headers=['drift x [nm]'], save=True, index=False, agg='mean')
    def dx(self:np.float32) :
        return "nodrift"

    @column(headers=['drift y [nm]'], save=True, index=False, agg='mean')
    def dy(self:np.float32) :
        return "nodrift"

    @column(headers=['drift z [nm]'], save=True, index=False, agg='mean')
    def dz(self:np.float32) :
        return "nodrift"

    @column(headers=['drift phi [rad]'], save=True, index=False, agg='mean')
    def dphi(self:np.float32) :
        return "nodrift"

    @column(headers=['drift freq [Hz]'], save=True, index=False, agg='mean')
    def dfreq(self:np.float32) :
        return "nodrift"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-20
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : smlmLP
# Module        : Locs

"""
This class define objects corresponding to localizations sets for one experiment.
"""



# %% Libraries
from corelp import folder, selfkwargs, prop
from smlmlp import open_df, save_df, LocsDataFrame, DetsDataFrame, Config, columns
from pathlib import Path
import pandas as pd
import numpy as np



# %% Class
class Locs(LocsDataFrame) :
    '''
    This class define objects corresponding to localizations sets for one experiment.

    Parameters
    ----------
    a : int or float
        TODO.

    Attributes
    ----------
    _attr : int or float
        TODO.

    Examples
    --------
    >>> from smlmlp import Locs
    ...
    >>> instance = Locs(TODO)
    '''


    # Creates Locs objects
    def __new__(cls, source=None, /, **kwargs) :

        #If Locs object is given as input returns it directly
        if isinstance(source, cls) :
            return source

        #Creates a new Locs object
        return object.__new__(cls)



    # Initialize with any object
    df_dict = None
    def __init__(self, source=None, config=None) :
        self.df_dict = dict(detections=DetsDataFrame(self))

        #Changes attributes from kwargs values
        if isinstance(source, Locs) :
            return #break if already a Locs object

        #Opening data
        if source is not None :
            self.open(source)

        # Config
        if config is not None :
            self.config_source = config
        self.config = Config(config=self.config_source, **self.config_kwargs)

    # Config
    @prop()
    def config_source(self) :
        return None
    @config_source.setter
    def config_source(self, value) :
        if getattr(self, '_config_source', None) is not None :
            raise ValueError('Config source is defined twice')
        self._config_source = value
    @property
    def config_kwargs(self) :
        return dict(nfiles=1) if self.config_source is None else dict()

    # Detections properties
    @property
    def detections(self) :
        return self.df_dict['detections']
    @property
    def ndetections(self) :
        return len(self.detections)


    def filter(self, *filter_names, mask=None) :
        filters = [getattr(self, name, None) for name in filter_names]
        if mask is not None : filters += [mask]
        mask = np.logical_and.reduce(filters)
        self.keep = mask
        df_list = [df[mask].drop(columns=[columns[name].header for name in filter_names if columns[name].header in df]) for df in self.df_dict.values()]
        return Locs(df_list, config=self.config)



    def open(self, source) :
        if source is None :
            return
        elif isinstance(source, pd.DataFrame) :
            open_df(self, source)
        elif isinstance(source, dict) :
            selfkwargs(self, source)
        elif isinstance(source, str) or isinstance(source,Path('').__class__) :
            path = Path(source)
            if path.is_dir() : # folder : open csv and json
                stem = path.stem
                detection_file = path / f'{stem}_[detections].csv'
                if not detection_file.exists() :
                    raise ValueError('Cannot open foler without "detections" file')
                self.open(detection_file)
                for file in path.glob(f'{stem}_[*].csv') :
                    if file == detection_file : continue
                    self.open(file)
                config_file = path / f'{stem}_[metadata].json'
                if config_file.exists() :
                    self.open(config_file)
            else :
                if path.suffix == '.csv' :
                    self.open(pd.read_csv(path))
                elif path.suffix == '.json' :
                    self.config_source = path
                else :
                    raise SyntaxError(f'Locs cannot open file {path}')
        elif isinstance(source, list) or isinstance(source, tuple) :
            for value in source :
                self.open(value)
        else :
            raise SyntaxError('source type was not recognized for Locs')



    def save(self, path, file=None) :
        path = Path(path) if file is None else Path(path) / file
        saving_folder = folder(path.with_suffix(''), warning=False)
        stem = saving_folder.stem
        mainpath = saving_folder / f'{stem}.csv'
        
        # Saving df
        for df_name, df in self.df_dict.items() :
            if df is None or len(df.columns) == 0 : continue
            path = mainpath.with_stem(f'{stem}_[{df_name}]')
            save_df(df, path, df.head2save)

        # Saving metadata
        mainpath = saving_folder / f'{stem}_[metadata].json'
        self.config.save(mainpath)



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import columns, DataFrame, MainDataFrame, LocsDataFrame
import numpy as np



# %% Class
class column() :
    '''
    This object is a decorator to apply to localization columns.
    
    Parameters
    ----------
    a : int or float
        TODO.

    Attributes
    ----------
    _attr : int or float
        TODO.

    Examples
    --------
    >>> from smlmlp import column
    ...
    >>> instance = column(TODO)
    '''



    def __init__(self, *, headers, save=True, index=False, agg='mean'):
        ''' First called, bracket of decorator '''
        self.headers = headers #List of possible headers, first is default one
        self.header = headers[0]
        self.save = save #True to save this column in files
        self.index = index # True if index is the index of the class
        self.agg = agg # Aggregation method



    def __call__(self, func) :
        ''' Called after initialization '''

        #Get column info
        self.func = func #Default behavior function
        self.col = func.__name__ #column shortcut name
        self.dtype = func.__annotations__['self'] # dtype of column

        #Updating columns dictionnaries
        if self.col in columns: raise SyntaxError(f"Column {self.col} is defined twice")
        columns[self.col] = self
        for header in self.headers :
            if header in columns.headers: raise SyntaxError(f"Header {header} is defined twice")
            columns.headers[header] = self

        return self



    def __set_name__(self, cls, name):
        ''' Called when assigned to a class '''
        self.cls = cls # dataframe object
        if self.index :
            self.df_name = f'{self.header}s'
            if self.df_name != self.cls.__name__ :
                raise SyntaxError(f'Index column {self.col} does not coincide with DataFrame name {self.cls.__name__}')
            self.cls.index_name = self.header

        # Assign column dict
        if not hasattr(self.cls, "columns_list") : self.cls.columns_dict = {}
        self.cls.columns_dict[self.col] = self

        # Assign saving list
        if self.index :
            for header in self.headers :
                MainDataFrame.head2save.append(header)
        else :
            if not hasattr(self.cls, "head2save") : self.cls.head2save = []
            for header in self.headers :
                self.cls.head2save.append(header)
        

        # Get column object from dataframe instance
        @property
        def _col(instance) :
            return self
        setattr(self.cls, f'_{self.col}', _col)

        # Set on MainDataFrame
        if issubclass(self.cls, MainDataFrame) :
            @property
            def merged_col(df) :
                if self.header not in df.columns :
                    dets = df.locs.detections
                    df[self.header] = dets.groupby(df.index.name)[self.header].agg(self.agg).to_numpy()
                return df[self.header].to_numpy()
            if hasattr(DataFrame, self.col) : raise SyntaxError(f'{self.col} cannot be defined twice in DataFrame')
            setattr(DataFrame, self.col, merged_col)
        
        # Set on DataFrame
        else :
            if self.index :
                @property
                def index_col(dets) :
                    if self.header not in dets.columns :
                        return None
                    return dets[self.header].to_numpy()
                @index_col.setter
                def index_col(dets, value) :
                    if value is None :
                        dets.drop(columns=[self.header], inplace=True)
                        dets.df_dict.pop(self.df_name)
                    else :
                        array = np.asarray(value, dtype=self.dtype)
                        dets[self.header] = array
                if hasattr(MainDataFrame, self.col) : raise SyntaxError(f'{self.col} cannot be defined twice in MainDataFrame')
                setattr(MainDataFrame, self.col, index_col)
                @property
                def get_df(locs) :
                    if self.df_name not in locs.df_dict :
                        from smlmlp import dataframes
                        locs.df_dict[self.df_name] = dataframes[self.df_name](locs)
                    return locs.df_dict[self.df_name]
                if hasattr(LocsDataFrame, self.df_name) : raise SyntaxError(f'{self.df_name} cannot be defined twice in LocsDataFrame')
                setattr(LocsDataFrame, self.df_name, get_df)
                @property
                def len_df(locs) :
                    df = getattr(locs, self.df_name)
                    return len(df)
                if hasattr(LocsDataFrame, f'n{self.df_name}') : raise SyntaxError(f'n{self.df_name} cannot be defined twice in LocsDataFrame')
                setattr(LocsDataFrame, f'n{self.df_name}', len_df)
            else :
                @property
                def spread_col(dets) :
                    if self.header not in dets.columns :
                        df = getattr(dets.locs, self.cls.__name__)
                        dets[self.header] = dets[df.index.name].map(df[self.header]).to_numpy()
                    return dets[self.header].to_numpy()
                if hasattr(MainDataFrame, self.col) : raise SyntaxError(f'{self.col} cannot be defined twice in LocsDataFrame')
                setattr(MainDataFrame, self.col, spread_col)
                


    def __get__(self, instance, cls):
        # Good practice getter
        if instance is None:
            return self

        # Index
        if self.index :
            return instance.index.to_numpy()

        # Gets from dataframe
        if self.header in instance.columns :
            return instance[self.header].to_numpy()

        # Automatic calculation
        newcol = self.func(instance)
        if newcol is None :
            return None

        # Substitute
        if isinstance(newcol, str) :
            return getattr(instance, newcol)
        
        # Calculated
        setattr(instance, self.col, newcol)
        return getattr(instance, self.col)



    def __set__(self, instance, value):
        
        # Index
        if self.index :
            if value is None : raise SyntaxError('Setting Dataframe index to None is not possible')
            return instance.index.to_numpy()

        #Removing column from dataframe (value=None)
        if value is None :
            instance.drop(columns=[self.header], inplace=True)

        #Setting column
        array = np.asarray(value, dtype=self.dtype)
        instance[self.header] = array




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-25
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : smlmLP
# Module        : Config

"""
This class stores configuration values in an object that can be saved and loaded.
"""



# %% Libraries
from corelp import prop, selfkwargs, folder
from pathlib import Path
import json
import numpy as np
from smlmlp import metadatum, Camera, Channel
from contextlib import ExitStack
import tifffile as tiff
from stacklp import shapetif
from arrlp import coordinates



# %% Class
class Config() :
    '''
    This class stores configuration values in an object that can be saved and loaded.
    
    Parameters
    ----------
    a : int or float
        TODO.

    Attributes
    ----------
    _attr : int or float
        TODO.

    Examples
    --------
    >>> from smlmlp import Config
    ...
    >>> instance = Config(TODO)
    '''


    # Init
    def __init__(self, *tif_paths, config=None, **kwargs) :

        # Opening tif files
        self.nfiles = len(tif_paths)

        # Loading previous configuration
        if config is not None :
            if isinstance(config, Config) :
                config = config.metadata
            if isinstance(config, str) or isinstance(config,Path('').__class__):
                config_file = Path(config).with_suffix('.json')
                if config_file.exists() :
                    with open(config_file, "r") as file:
                        config = json.load(file)
                    config_folder = config_file.parent / "_config_data"
                    if config_folder.exists() :
                        data = {}
                        for file in config_folder.glob('*.npy') :
                            key = file.stem
                            data[key] = np.load(file)
                        config["Data"] = data
                else :
                    raise SyntaxError(f'config path was not recognized: {config}')
            for group_dict in config.values() :
                selfkwargs(self, group_dict)
        selfkwargs(self, kwargs)

        # Opening files
        if len(tif_paths) :
            with ExitStack() as stack :
                tifs = [stack.enter_context(tiff.TiffFile(file)) for file in tif_paths]
                shapes = [shapetif(tif) for tif in tifs]

            # Check number of frames
            for shape in shapes :
                if shape[0] != shapes[0][0] :
                    raise ValueError('All tiff files do not have the same number of frames which is not possible for a single SMLM acquisition')
            self.nframes = shapes[0][0]
            self.npixels = [shape[1:3] for shape in shapes]



    @property
    def metadata(self) :
        data = {}
        for group_name, group_list in metadatum.groups.items() :
            data_dict = {}
            data[group_name] = data_dict
            for datum in group_list :
                value = getattr(self, f'_{datum}', None)
                if value is not None :
                    data_dict[datum] = json_convert(value)
        return data



    def save(self, path, file=None) :
        path = Path(path)
        if file is not None : path = path / file
        path = path.with_suffix('.json')
        metadata = self.metadata
        data = metadata.pop('Data')
        with open(path, 'w') as json_file:
            json.dump(metadata, json_file, indent=4)
        config_folder = folder(path.parent / "_config_data", warning=False)
        for key, value in data.items() :
            np.save(config_folder / f"{key}.npy", np.asarray(value))




    # --- Below are the metadata, organized by group---



    # Cameras

    @metadatum('Cameras')
    def nfiles(self) :
        raise SyntaxError('Should not be called')
    @nfiles.setter
    def nfiles(self, value) :
        self.cameras = [Camera(self) for _ in range(int(value))]
    @property
    def _nfiles(self) :
        return len(self.cameras)

    @property
    def FOV_max(self) :
        ymin = min([camera.FOV_max[0] for camera in self.cameras])
        xmin = min([camera.FOV_max[1] for camera in self.cameras])
        return ymin, xmin

    @property
    def bbox(self) :
        FOV = self.FOV_max
        return [camera.FOV2bbox(FOV) for camera in self.cameras]
    @property
    def frame_bytes(self) : # gigabytes/frame
        return sum([camera.frame_bytes for camera in self.cameras])



    # Channels

    @property
    def total_nchannels(self) :
        return sum(self.nchannels)

    @property
    def channels(self) :
        return [channel for camera in self.cameras for channel in camera.channels]



    # Loads

    @metadatum('Loads', dtype=int)
    def nframes(self) :
        return 60000



    # Time

    @metadatum('Cameras', dtype=float)
    def exposure(self) : # [ms]
        return 50.

    @metadatum('Blinks')
    def on_time(self) : # ms
        return 50.



    # Background configurations

    @metadatum('Backgrounds', dtype=float)
    def median_window(self) : # For temporal median [ms]
        return self.on_time * 10

    @metadatum('Backgrounds', dtype=bool)
    def do_temporal_median(self) : # For temporal median
        return True

    @metadatum('Backgrounds', dtype=float)
    def mean_radius(self) : # For spatial mean [nm]
        return max(self.psf_sigma) * 8

    @metadatum('Backgrounds', dtype=bool)
    def do_spatial_mean(self) : # For spatial mean
        return True

    @metadatum('Backgrounds', dtype=float)
    def opening_radius(self) : # For spatial opening [nm]
        return max(self.psf_sigma)

    @metadatum('Backgrounds', dtype=bool)
    def do_spatial_opening(self) : # For spatial opening
        return False



    # Signal configurations

    @metadatum('Signals', dtype=bool)
    def do_spatial_filter(self) : # For spatial filter
        return True
    @metadatum('Signals', dtype=bool)
    def do_temporal_filter(self) : # For temporal filter
        return False

    @metadatum('Signals')
    def spatial_subtract_factor(self) :
        return 2.5
    @spatial_subtract_factor.setter
    def spatial_subtract_factor(self, value) :
        self._spatial_subtract_factor = value
        for channel in self.channels:
            channel._spatial_kernel = None

    @metadatum('Signals')
    def temporal_subtract_factor(self) :
        return 0.
    @temporal_subtract_factor.setter
    def temporal_subtract_factor(self, value) :
        self._temporal_subtract_factor = value
        self._temporal_kernel = None

    @property
    def temporal_kernel_shape(self) :
        on_frames = self.on_time / self.exposure
        if self.temporal_subtract_factor > 1 : on_frames *= self.temporal_subtract_factor
        on_frames = int(np.ceil(on_frames))
        return (on_frames * 10 + 1),

    @property
    def on_time_kernel(self) :
        T, = coordinates(shape=self.temporal_kernel_shape, pixel=self.exposure)
        from funclp import Exponential1
        exponential = Exponential1(tau=self.on_time)
        k = exponential(np.abs(T))
        k /= k.sum()
        return k.astype(np.float32)

    @property
    def temporal_subtract_kernel(self) :
        if self.temporal_subtract_factor <= 1 : return None
        exposure = self.exposure / self.temporal_subtract_factor
        T, = coordinates(shape=self.temporal_kernel_shape, pixel=exposure)
        from funclp import Exponential1
        exponential = Exponential1(tau=self.on_time)
        k = exponential(np.abs(T))
        k /= k.sum()
        return k.astype(np.float32)

    @prop(cache=True)
    def temporal_kernel(self) :
        if self.temporal_subtract_factor <= 1 :
            return self.on_time_kernel
        k = self.on_time_kernel - self.temporal_subtract_kernel
        return k - k.mean()



    # Detection
    @metadatum('Detections')
    def snr_thresh(self) :
        return 4.



# Adding Camera metadata
for data, group in Camera.metadata :
    @metadatum(group, name=data)
    def camera_property(self, data=data) :
        return [getattr(camera, data) for camera in self.cameras]
    @camera_property.setter
    def camera_property(self, value, data=data) :
        try :
            if len(value) != self.nfiles : raise ValueError('Value set does not have same number of elements as cameras')
        except TypeError :
            value = [value for _ in range(self.nfiles)]            
        for camera, v in zip(self.cameras, value) :
            setattr(camera, data, v)
    setattr(Config, data, camera_property)
    @property
    def _camera_property(self, data=data) :
        value = [getattr(camera, f'_{data}', None) for camera in self.cameras]
        bool_ = [v is None for v in value]
        return None if any(bool_) else value
    setattr(Config, f'_{data}', _camera_property)
    @property
    def channel_property(self, data=data) :
        return [getattr(channel, data) for channel in self.channels]
    setattr(Config, f'channel_{data}', channel_property)

# Adding Camera properties
for data in Camera.properties :
    @property
    def camera_property(self, data=data) :
        return [getattr(camera, data) for camera in self.cameras]
    @camera_property.setter
    def camera_property(self, value, data=data) :
        try :
            if len(value) != self.nfiles : raise ValueError('Value set does not have same number of elements as cameras')
        except TypeError :
            value = [value for _ in range(self.nfiles)]            
        for camera, v in zip(self.cameras, value) :
            setattr(camera, data, v)
    setattr(Config, data, camera_property)
    @property
    def channel_property(self, data=data) :
        return [getattr(channel, data) for channel in self.channels]
    setattr(Config, f'channel_{data}', channel_property)



# Adding Channel metadata
for data, group in Channel.metadata :
    @metadatum(group, name=data)
    def channel_property(self, data=data) :
        return [getattr(channel, data) for channel in self.channels]
    @channel_property.setter
    def channel_property(self, value, data=data) :
        try :
            if len(value) != self.total_nchannels : raise ValueError('Value set does not have same number of elements as channels')
        except TypeError :
            value = [value for _ in range(self.total_nchannels)]     
        for channel, v in zip(self.channels, value) :
            setattr(channel, data, v)
    setattr(Config, data, channel_property)
    @property
    def _channel_property(self, data=data) :
        value = [getattr(channel, f'_{data}', None) for channel in self.channels]
        bool_ = [v is None for v in value]
        return None if any(bool_) else value
    setattr(Config, f'_{data}', _channel_property)

# Adding Channel properties
for data in Channel.properties :
    @property
    def channel_property(self, data=data) :
        return [getattr(channel, data) for channel in self.channels]
    @channel_property.setter
    def channel_property(self, value, data=data) :
        try :
            if len(value) != self.total_nchannels : raise ValueError('Value set does not have same number of elements as channels')
        except TypeError :
            value = [value for _ in range(self.total_nchannels)]     
        for channel, v in zip(self.channels, value) :
            setattr(channel, data, v)
    setattr(Config, data, channel_property)





# Convert to json function
def json_convert(value) :
    if isinstance(value, bool) or isinstance(value, np.bool_) : value = bool(value)
    if isinstance(value, int) or isinstance(value, np.integer) : value = int(value)
    if isinstance(value, float) or isinstance(value, np.floating) : value = float(value)
    if isinstance(value, tuple) or isinstance(value, list) : value = [json_convert(v) for v in value]
    if isinstance(value, np.ndarray) :
        value = value.ravel()
        if np.issubdtype(value.dtype, np.bool_) :
            value = value.astype(bool)
        elif np.issubdtype(value.dtype, np.integer) :
            value = value.astype(int)
        elif np.issubdtype(value.dtype, np.floating) :
            value = value.astype(float)
        value = list(value)
    return value



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-23
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : smlmLP
# Module        : dataframes

"""
This is a dict of all the dataframes that are created.
"""



# %% Libraries
from pathlib import Path
import importlib



path = Path(__file__).parent / f'_functions/dataframes'
names = [file.stem[1:] for file in path.iterdir() if file.suffix == '.py' and not file.name.startswith('__')]
dataframes = {name: getattr(importlib.import_module(f'._functions.dataframes._{name}', package=__package__), name) for name in names}



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from smlmlp import block, Config
from arrlp import gc, get_xp, img_gaussianfilter
from funclp import Exponential1
from stacklp import temporal_autocorr as stack_autocorr
import numpy as np
from scipy.optimize import curve_fit



# %% Function
@block()
def blink_temporal_on(channels, /, crop=None, psf_sigma=1., *, exposure=1., channel_pixel=1., cuda=False, parallel=False) :
    '''
    This function creates the temporal autocorrelation for on time measurements.
    '''

    # Optimization parameters
    xp = get_xp(cuda)

    # Get pixel
    pixel = Config(nfiles=len(channels), pixel=channel_pixel).pixel
    
    # Get coordinates
    if crop is None: crop = int(len(channels) // 2 - 1)
    T = np.arange(crop)
    t = T * exposure

    results = {'ac': []}
    for i, (channel, pix) in enumerate(zip(channels, pixel)) :

        # Calculating autocorrelation
        gc()
        bkgd = img_gaussianfilter(channel, sigma=psf_sigma[i] * 3, pixel=pix, cuda=cuda, parallel=parallel)
        bkgd = xp.minimum(bkgd, channel)
        channel = channel - bkgd
        f0 = int(channel.shape[0]//2)
        ac = stack_autocorr(channel, cuda=cuda, parallel=parallel)[f0 + 1: f0 + 1 + crop]
        ac = ac.mean(axis=(1,2))
        ac -= ac[-1]
        ac /= ac[0]
        if cuda : ac = xp.asnumpy(ac)
        results['ac'].append(ac)

    # Averaging
    gc()
    y = np.zeros_like(results['ac'][0])
    for ac in results['ac'] :
        y += (ac / len(results['ac']))
    results['average'] = y

    # Fitting
    p0 = [1., -0.5] # tau, offset
    bounds = ([0., -1.], [len(y), 1.]) # (lower, upper)
    expodecay = Exponential1()
    func2fit = lambda t, tau, offset: expodecay(t, tau=tau, offset=offset)
    popt, _ = curve_fit(func2fit, T, y, p0=p0, bounds=bounds)
    tau, offset = popt
    results['fit'] = func2fit(T, *popt)
    results['time'] = t
    on_time = tau * exposure # ms

    return on_time, results